#include "settings.h"
#include <Arduino.h>
#include <ArduinoJson.h>
#include <AsyncJson.h>
#include <DHT.h>
#include <EEPROM.h>
#include <ESPAsyncWebServer.h>
#include <SPIFFS.h>
#include <WiFi.h>

void readSensor();
void setupSensor();

const char *ssid = "Vegova RVP4";
const char *password = "Vegova_RVP4";

DHT *dht;
Settings settings;

AsyncWebServer server(80);
AsyncEventSource events("/events");

void setupWifi() {
  Serial.println("Connecting to WiFi");
  WiFi.begin(ssid, password);

  while (WiFi.status() != WL_CONNECTED) {
    delay(1000);
    Serial.print(".");
  }
  Serial.println();
  Serial.print("Connected: ");
  Serial.println(WiFi.localIP());
}

void onConnect(AsyncEventSourceClient *client) {
  Serial.println("New client connected");
  readSensor();
}

void setupServer() {
  if (!SPIFFS.begin(true)) {
    Serial.println("Failed to mount file system");
    return;
  }

  // POST /set - set the output pin
  AsyncCallbackJsonWebHandler *setHandler = new AsyncCallbackJsonWebHandler(
      "/set", [](AsyncWebServerRequest *req, JsonVariant &json) {
        if (json["pin"].isNull() || json["value"].isNull()) {
          req->send(400);
          return;
        }
        digitalWrite(json["pin"].as<int>(), json["value"].as<bool>());
        readSensor();
        req->send(200);
      });
  server.addHandler(setHandler);

  // GET /settings - Get the current settings
  server.on("/settings", HTTP_GET, [](AsyncWebServerRequest *req) {
    StaticJsonDocument<200> json;
    json["refreshInterval"] = settings.getRefreshInterval();
    json["sensorPin"] = settings.getSensorPin();
    json["onTemp"] = settings.getOnTemp();
    json["onHum"] = settings.getOnHum();
    json["tempOffset"] = settings.getTempOffset();
    json["humOffset"] = settings.getHumOffset();
    json["heaterOn"] = settings.getHeaterOn();
    json["dehumOn"] = settings.getDehumOn();

    String buffer;
    serializeJson(json, buffer);

    req->send(200, "application/json", buffer);
  });

  // PATCH /settings - Update settings
  AsyncCallbackJsonWebHandler *settingsHandler =
      new AsyncCallbackJsonWebHandler(
          "/settings", [](AsyncWebServerRequest *req, JsonVariant &json) {
            if (!json["refreshInterval"].isNull()) {
              settings.setRefreshInterval(json["refreshInterval"].as<long>());
            }
            if (!json["sensorPin"].isNull()) {
              settings.setSensorPin(json["sensorPin"].as<unsigned int>());
              setupSensor();
            }
            if (!json["onTemp"].isNull()) {
              settings.setOnTemp(json["onTemp"].as<float>());
            }
            if (!json["onHum"].isNull()) {
              settings.setOnHum(json["onHum"].as<float>());
            }
            if (!json["tempOffset"].isNull()) {
              settings.setTempOffset(json["tempOffset"].as<float>());
            }
            if (!json["humOffset"].isNull()) {
              settings.setHumOffset(json["humOffset"].as<float>());
            }
            req->send(200);
          });
  server.addHandler(settingsHandler);

  server.onNotFound([](AsyncWebServerRequest *request) {
    if (request->method() == HTTP_OPTIONS) {
      request->send(200);
    } else {
      request->send(404);
    }
  });

  events.onConnect(onConnect);
  server.addHandler(&events);
  server.serveStatic("/", SPIFFS, "/").setDefaultFile("index.html");
  server.begin();

  Serial.println("Server running");
}

void setupSensor() {
  dht = new DHT(settings.getSensorPin(), DHT11);
  dht->begin();
}

void readSensor() {
  float temp = dht->readTemperature();
  float hum = dht->readHumidity();

  float onTemp = settings.getOnTemp();
  float offTemp = onTemp + settings.getTempOffset();

  float onHum = settings.getOnHum();
  float offHum = onHum + settings.getHumOffset();

  if (temp <= onTemp) {
    digitalWrite(16, HIGH);
  } else {
    digitalWrite(16, LOW);
  }

  if (hum >= onHum) {
    digitalWrite(17, HIGH);
  } else {
    digitalWrite(17, LOW);
  }

  StaticJsonDocument<200> json;
  json["temp"] = (int)(temp * 100 + 0.5) / 100.0;
  json["hum"] = hum;
  json["out1"] = (bool)digitalRead(16);
  json["out2"] = (bool)digitalRead(17);

  String buffer;
  serializeJson(json, buffer);

  events.send(buffer.c_str(), "message", millis());
}

void setup() {
  Serial.begin(115200);

  pinMode(16, OUTPUT);
  pinMode(17, OUTPUT);

  setupWifi();
  setupServer();

  settings.load();

  setupSensor();
}

unsigned long prevMillis = 0;

void loop() {
  unsigned long currMillis = millis();
  if (currMillis - prevMillis > settings.getRefreshInterval()) {
    prevMillis = currMillis;
    readSensor();
  }
}